/*
 * edit.c
 * line editing functions for powwow
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/time.h>

#include "powwow.h"
#include "edit.h"
#include "tcp.h"
#include "tty.h"
#include "eval.h"

edit_function internal_functions[] = {
  {NULL, (void (*)()) NULL, },
  {"&enter-line", enter_line, },
  {"&complete-word", complete_word, },
  {"&complete-line", complete_line, },
  {"&del-char-left", del_char_left, },
  {"&del-char-right",  del_char_right, },
  {"&prev-char", prev_char, },
  {"&prev-line", prev_line, },
  {"&next-char", next_char, },
  {"&next-line", next_line, },
  {"&to-history", enter_into_history, },
  {"&clear-line", clear_line, },
  {"&begin-of-line", beginning_of_line, },
  {"&end-of-line", end_of_line, },
  {"&kill-to-eol", kill_to_eol, },
  {"&transpose", transpose_chars, },
  {"&transpose-words", transpose_words, },
  {"&redraw-line", refresh_input_line, },
  {"&suspend", suspend_powwow, },
  {"&del-word-left", del_word_left, },
  {"&del-word-right", del_word_right, },
  {"&prev-word", prev_word, },
  {"&upcase-word", upcase_word, },
  {"&downcase-word", downcase_word, },
  {"&next-word", next_word, },
  {NULL, (void (*)()) NULL }
};


int lookup_edit_name(name)
char *name;
{
  int i;
  char *fname;

  for (i=1; (fname = internal_functions[i].name); i++)
    if (!strcmp(name, fname))
      return i;

  return 0;
}

int lookup_edit_function(funct)
void (*funct)();
{
  int i;
  void (*ffunct)();

  for (i=1; (ffunct = internal_functions[i].funct); i++)
    if (funct == ffunct)
      return i;

  return 0;
}

/*
 * redraw the input line
 */
void refresh_input_line()
{
    tty_clear_input_line(0);
    col0 = promptlen = promptstr[0] = surely_isprompt = '\0'; 
    tty_putc('\n');
    if (line0 < lines - 1)
      line0 ++;
    tty_draw_input_line();
}

/*
 * GH: transpose two words to the left
 */
void transpose_words()
{
    /* other refers to the word to the left, this is the one we are at */

    int this_so, other_so, this_eo, other_eo;
    char *delims = WORDDELIM;
    char buf[LINELEN];
    int n;

    if (pos > 2) {

	this_eo = this_so = pos;
	/* optionally traceback to find a word */
	while (this_so && strchr(delims, edbuf[this_so]))
	    this_so--;

	/* now find where the current word ends */
	while (this_eo < edlen && !strchr(delims, edbuf[this_eo]))
	    this_eo++;

	/* found a word; now find its start */
	while (this_so > 0 && !strchr(delims, edbuf[this_so - 1]))
	    this_so--;

	if (this_so < 2)
	    return;		/* impossible that there's another word */

	other_so = this_so - 1;
	while (other_so >= 0 && strchr(delims, edbuf[other_so]))
	    other_so--;
	if (other_so < 0)
	    return;
	other_eo = other_so + 1;
	while (other_so > 0 && !strchr(delims, edbuf[other_so - 1]))
	    other_so--;

	sprintf(buf, "%.*s%.*s%.*s",
		this_eo - this_so, edbuf + this_so, 
		this_so - other_eo, edbuf + other_eo,
		other_eo - other_so, edbuf + other_so);

	tty_moveto(other_so);
	for (n = 0; buf[n]; tty_overtype_follow(buf[n++]));
    }	
}

/*
 * transpose two characters to the left
 */
void transpose_chars()
{
    int i, j;
    char c;
    if (pos > 1 || (pos > 0 && pos < edlen)) {
        if (pos < edlen) {
	  j = pos;
	  i = pos - 1;
	} else {
	  j = pos - 1;
	  i = pos - 2;
	}
	c = edbuf[j]; edbuf[j] = edbuf[i]; edbuf[i] = c;
	tty_gotoxy_opt(CURCOL(pos), CURLINE(pos), CURCOL(i), CURLINE(i));
	tty_putc(edbuf[i]);
	tty_gotoxy_opt(CURCOL(i+1), CURLINE(i+1), CURCOL(j), CURLINE(j));
	tty_putc(edbuf[j]);
	if (pos < edlen) {
	    pos++;
	    tty_gotoxy_opt(CURCOL(j+1), CURLINE(j+1), CURCOL(pos), CURLINE(pos));
	}
    }
}

/*
 * erase everything to the end of line
 */
void kill_to_eol()
{
    tty_puts (tty_clreoln);
    if (CURLINE(edlen) > CURLINE(pos)) {
	tty_printf("\n%s", tty_clreoscr);
	tty_gotoxy_opt(0, CURLINE(pos) + 1, CURCOL(pos), CURLINE(pos));
    }
    edbuf[pos] = '\0';
    edlen = pos;
}

/*
 * move cursor to end of line
 */
void end_of_line()
{
    tty_gotoxy_opt(CURCOL(pos), CURLINE(pos), CURCOL(edlen), CURLINE(edlen));
    pos = edlen;
}

/*
 * move cursor to beginning of line
 */
void beginning_of_line()
{
    tty_gotoxy_opt(CURCOL(pos), CURLINE(pos), col0, line0);
    pos = 0;
}

/*
 * delete a character to the right
 */
void del_char_right()
{
    if (pos < edlen)
	tty_delete_nofollow_chars(1);
}

/*
 * delete a character to the left
 */
void del_char_left()
{
    if (pos) {
	pos--;
	tty_gotoxy_opt(CURCOL(pos + 1), CURLINE(pos + 1), CURCOL(pos), CURLINE(pos));
	tty_delete_nofollow_chars(1);
    }
}

/*
 * enter a line into history, but don't do anything else
 */
void enter_into_history()
{
    if (!edlen)
      return;
    tty_clear_input_line(0);
    put_in_history(edbuf);
    pickline = curline;
    *edbuf = '\0';
    pos = edlen = 0;
}

/*
 * clear input line, but do nothing else
 */
void clear_line()
{
    if (!edlen)
      return;
    tty_clear_input_line(0);
    pickline = curline;
    *edbuf = '\0';
    pos = edlen = 0;
}

/*
 * enter a line
 */
void enter_line()
{
    char *p;

    tty_gotoxy_opt(CURCOL(pos), CURLINE(pos), CURCOL(edlen), CURLINE(edlen));
    tty_printf("%s\n", edattrend);

    line0 = CURLINE(edlen);
    if (line0 < lines - 1) line0++;
#ifdef NEWBIE
    denewbiefy();
#endif
    if (recordfile)
        fprintf(recordfile, "%s\n", edbuf);

    col0 = 0;
    error = internal = 0;

    if (!*edbuf || (verbatim && *edbuf != '#'))
	tcp_write(edbuf);
    else
	parse_user_input(edbuf, 1);

    if (internal >= 0) {
	showprompt();
	col0 = printstrlen(promptstr);
	promptlen = strlen(promptstr);
    } else if (internal != -2)
	col0 = promptlen = promptstr[0] = surely_isprompt = '\0'; 

    /* don't put identical lines in history, nor short ones */
    p = hist[curline ? curline - 1 : MAXHIST - 1];
    if (!history_done && (!p || (edlen > 1 && strcmp(edbuf, p))))
	put_in_history(edbuf);
    history_done = 0;
    pickline = curline;
    if (inserted_next[0]) {
	strcpy(edbuf, inserted_next);
	inserted_next[0] = '\0';
    } else
	strcpy(edbuf, prefixstr);
    pos = edlen = strlen(edbuf);
    tty_draw_input_line();
}

/*
 * move one word forward
 */
void next_word()
{
    int i;
    for (i = pos; !isalnum(edbuf[i]) && edbuf[i]; i++);
    while (isalnum(edbuf[i])) i++;
    tty_gotoxy_opt(CURCOL(pos), CURLINE(pos), CURCOL(i), CURLINE(i));
    pos = i;
}

/*
 * move one word backward
 */
void prev_word()
{
    int i;
    for (i = pos; i && !isalnum(edbuf[i - 1]); i--);
    while (i && isalnum(edbuf[i - 1])) i--;
    tty_gotoxy_opt(CURCOL(pos), CURLINE(pos), CURCOL(i), CURLINE(i));
    pos = i;
}

/*
 * delete word to the right
 */
void del_word_right()
{
    int i;
    for (i = pos; !isalnum(edbuf[i]) && i < edlen; i++);
    while (isalnum(edbuf[i])) i++;
    tty_delete_nofollow_chars(i - pos);
}

/*
 * delete word to the left
 */
void del_word_left()
{
    int i;
    for (i = pos; i && !isalnum(edbuf[i - 1]); i--);
    while (i && isalnum(edbuf[i - 1])) i--;
    tty_gotoxy_opt(CURCOL(pos), CURLINE(pos), CURCOL(i), CURLINE(i));
    i = pos - i; pos -= i;
    tty_delete_nofollow_chars(i);
}

/*
 * GH: make word upcase
 */
void upcase_word()
{
    int i = pos, opos = pos;

    if (last_edit_cmd == upcase_word)
	i = 0;
    else {
	while (i > 0 && ISWORDDELIM(edbuf[i])) i--;
	while (i > 0 && !ISWORDDELIM(edbuf[i - 1])) i--;
    }
    tty_moveto(i);
    while (!ISWORDDELIM(edbuf[pos]) ||
	  (last_edit_cmd == upcase_word && edbuf[pos])) {
	tty_overtype_follow(edbuf[pos] = toupper(edbuf[pos]));
	i++;
    }
    tty_moveto(opos);
}

/*
 * GH: make word downcase
 */
void downcase_word()
{
    int i = pos, opos = pos;

    if (last_edit_cmd == downcase_word)
	i = 0;
    else {
	while (i > 0 && ISWORDDELIM(edbuf[i])) i--;
	while (i > 0 && !ISWORDDELIM(edbuf[i - 1])) i--;
    }
    tty_moveto(i);
    while (!ISWORDDELIM(edbuf[pos]) ||
	  (last_edit_cmd == downcase_word && edbuf[pos])) {
	tty_overtype_follow(edbuf[pos] = tolower(edbuf[pos]));
	i++;
    }
    tty_moveto(opos);
}

/*
 * get previous line from history list
 */
void prev_line()
{
    int i = pickline - 1;
    if (i < 0) i = MAXHIST - 1;
    if (hist[i]) {
	if (hist[pickline] && strcmp(hist[pickline], edbuf)) {
	  free(hist[pickline]);
	  hist[pickline] = NULL;
	}
	if (!hist[pickline]) {
	  if (!(hist[pickline] = str_dup(edbuf))) {
	      errmsg("malloc");
	      return;
	  }
	}
	pickline = i;
	tty_clear_input_line(0);
	strcpy(edbuf, hist[pickline]);
	pos = edlen = strlen(edbuf);
	tty_draw_input_line();
    }
}

/*
 * get next line from history list
 */
void next_line()
{
    int i = pickline + 1;
    if (i == MAXHIST) i = 0;
    if (hist[i]) {
	if (hist[pickline] && strcmp(hist[pickline], edbuf)) {
	  free(hist[pickline]);
	  hist[pickline] = NULL;
	}
	if (!hist[pickline]) {
	  if (!(hist[pickline] = str_dup(edbuf))) {
	      errmsg("malloc");
	      return;
	  }
	}
	pickline = i;
	tty_clear_input_line(0);
	strcpy(edbuf, hist[pickline]);
	edlen = pos = strlen(edbuf);
	tty_draw_input_line();
    }
}

/*
 * move one char backward
 */
void prev_char()
{
    if (pos) {
	pos--;
	tty_gotoxy_opt(CURCOL(pos + 1), CURLINE(pos + 1), CURCOL(pos), CURLINE(pos));
    }
}

/*
 * move one char forward
 */
void next_char()
{
    if (pos < edlen) {
	pos++;
	tty_gotoxy_opt(CURCOL(pos - 1), CURLINE(pos - 1), CURCOL(pos), CURLINE(pos));
    }
}

/*
 * execute string as if typed
 */
void key_run_command(cmd)
char *cmd;
{
    tty_clear_input_line(opt_compact);

    /* GH: commented out this    tty_puts(edattrend); */
    if (echo_key)
        tty_puts(cmd);
    if (!opt_compact || echo_key)
        tty_putc('\n');

    col0 = 0;
    error = internal = 0;
    if (recordfile)
      fprintf(recordfile, "%s\n", cmd);

    parse_instruction(cmd, 1, 0);
    if (internal >= 0)
        showprompt();
    else if (internal != -2)
        promptlen = promptstr[0] = surely_isprompt = '\0'; 

    history_done = 0;
    tty_draw_input_line();
}
