/*
 * powwow.h
 * various constants etc
 */

#ifndef _POWWOW_H_
#define _POWWOW_H_

#if !defined(SYS_TIME_H) && !defined(_H_SYS_TIME)
#  include <sys/time.h>
#endif

typedef struct timeval vtime;   /* needs #include <sys/tyme.h> */

#ifdef AIX
#  include <sys/select.h>
#endif

#if !defined(STRING_H) && !defined(_H_STRING)
#  ifdef NO_BCOPY
#    define bcopy(a,b,c) memmove((b), (a), (c))
#  else /* not NO_BCOPY */
     void bcopy();
#  endif /* NO_BCOPY */
#  ifdef NO_BZERO 
#    define bzero(a,b) memset((a), 0, (b))
#  else /* not NO_BZERO */
     void bzero();
#  endif /* NO_BZERO */
#endif /* not !defined(STRING_H) && !defined(_H_STRING) */

#ifdef USE_REGEXP
#  include <regex.h>
#endif

#ifndef NULL
#  define NULL ((void *)0)
#endif

#ifndef LONG_MAX
#  define LONG_MAX ( (long) ((~0uL) >> 1) )
#endif

#ifndef LONG_MIN
#  define LONG_MIN ( (long) (((~0uL) >> 1) + 1) )
#endif

#define uSEC_PER_SEC 1000000L  /* microseconds in a second */
#define mSEC_PER_SEC 1000L     /* milliseconds in a second */
#define uSEC_PER_mSEC 1000L    /* microseconds in a millisecond */

#undef MIN2
#undef MAX2
#undef ABS
#undef SIGN
#define MIN2(a,b) ((a)<(b) ? (a) : (b))
#define MAX2(a,b) ((a)>(b) ? (a) : (b))
#define ABS(a)    ((a)> 0  ? (a) :(-a))
#define SIGN(a)   ((a)> 0  ?  1  : (a) ? -1 : 0)

/* macros to match parentheses */
#define ISRPAREN(c) ((c) == ')' || (c) == ']' || (c) == '}')
#define ISLPAREN(c) ((c) == '(' || (c) == '[' || (c) == '{')
#define LPAREN(c) ((c) == ')' ? '(' : ((c) == ']' ? '[' : '{'))

#define ISODIGIT(c) ((c) >= '0' && (c) <= '7')

#define PRINT_INT if(!internal) internal=1; tty_printf

#define MAXFDSCAN	32	/* max number of fd to scan in select() */
#define MAXHIST		128	/* number of history lines kept */
#define MIN_WORD_LEN	3	/* the minimum length for history words */
#define ESC		'\\'	/* special escape char */
#define ESC2		'`'	/* other special escape char */
#define CMDSEP		';'	/* command separator character */
#define DELIM		" ;"	/* word delimiting characters */
#define SPECIAL_CHARS	"{}();\"=" /* specials chars needing escape */
#define CAPLEN		20	/* max length of a terminal capability */
#define BUFSIZE		4096	/* general buffer size */
#define LINELEN		BUFSIZE	/* maximum line length */
#define PARAMLEN	1024	/* maximum length of strings (&x, $x) */
#define NUMPARAM	10	/* number of different params allowed */
#define NUMVAR		50	/* number of global variables allowed */
#define MAXMAPLEN	1000	/* maximum length of automapped path */
#define MAXWORDS	512	/* number of words kept for TAB-completion */
#define MPI		"~$#E"	/* MUME protocol introducer */
#define SOCKBUFSIZE	4096	/* socket buffer size for read */
#define MAXSUBOPT	256	/* max length of suboption string */
#define MAXARGS		16	/* max number of arguments to editor */
#define FLASHDELAY	500	/* time of parentheses flash in millisecs */
#define KBD_TIMEOUT	100	/* timeout for keyboard read in millisecs;
				   hope it's enough also for very slow lines */

#define MAX_STACK	100	/* maximum number of nested
                                 * action, alias, #for or #while */
#define MAX_LOOP	10000	/* maximum number of iterations in
                                 * for or #while */

#define ACTION_WEAK	0	/* GH: normal junk */
#define ACTION_REGEXP	1	/*     oh-so-mighty regexp */
#define ACTION_TYPES	(ACTION_REGEXP + 1)

#define DELIM_NORMAL	0	/* GH: normal word delimeters	*/
#define DELIM_PROGRAM	1	/*     ()[]{}.,;"'+/-*%		*/
#define DELIM_CUSTOM	2	/*     user-defined		*/
#define DELIM_MODES (DELIM_CUSTOM + 1)


/* macros to find cursor position from input buffer position */
#define CURLINE(pos) (((pos) + col0) / (cols - 1) + line0)
#define CURCOL(pos) (((pos) + col0) % (cols - 1))

#define ISMARKWILDCARD(c) ((c) == '&' || (c) == '$')

/* GH: the redefinable delimeters */
#define WORDDELIM (delimeters[word_delim_mode])
#define ISWORDDELIM(c) (strchr(WORDDELIM, (c)))

/*
 * Attribute codes: bit 0-4 for foreground color or effects, 5-9 for
 * background color. Color #16 is "none", so 0x0210 is "no attribute".
 */
#define COLORS         16         /* number of colors on HFT terminal */
#define LOWCOLORS      8          /* number of ANSI colors */
#define NO_COLOR       COLORS     /* no color change, use default */
#define BITS_COLOR     5          /* bits used for a color entry
                                     (16==none is a valid color) */
#define BITS_2COLOR    10         /* bits used for 2 color entries */
#define COLOR_MASK     0x1F       /* 5 (BITS_COLOR) bits set to 1, others 0 */

#define ATTR_BOLD      0x01
#define ATTR_BLINK     0x02
#define ATTR_UNDERLINE 0x04
#define ATTR_INVERSE   0x08

/*
 * WARNING: colors and attributes are currently using 14 bits.
 * type used is int and 0xFFFFFFFF ==  -1 is used
 * as 'invalid attribute' so in case ints are 16 bits,
 * there is only 1 bit left unused. In case ints are 32 bits, no problem.
 */

/* constructors / accessors for attribute codes */
#define ATTRCODE(attr, fg, bg) \
                (((attr) << BITS_2COLOR) | ((bg) << BITS_COLOR) | (fg))
#define FOREGROUND(attrcode) ((attrcode) & COLOR_MASK)
#define BACKGROUND(attrcode) (((attrcode) >> BITS_COLOR) & COLOR_MASK)
#define ATTR(attrcode)       ((attrcode) >> BITS_2COLOR)

#define NOATTRCODE ATTRCODE(0, NO_COLOR, NO_COLOR)

/*
 * NCSA telnet 2.2 doesn't reset the color when it receives "esc [ m",
 * so we must know what the normal colors are in order to reset it.
 * These colors can be changed with the #color command.
 */
#ifdef BUG_TELNET
#define DEFAULTFG 7		/* make default white text */
#define DEFAULTBG 4		/* on blue background */
#endif


/* generic linked list node (never actually created) */
typedef struct defnode {
    struct defnode *next;
    char *sortfield;
} defnode;

/* linked list nodes: keep the next pointer first, then string to sort by */
typedef struct aliasnode {
    struct aliasnode *next;
    char *name;
    char *subst;
} aliasnode;

typedef struct actionnode {
    struct actionnode *next;
    char *command, *label;
    int active;
    int type;				/* GH: allow regexp */
    char *pattern;
#ifdef USE_REGEXP
    regex_t *regexp;			/* 0 if type == ACTION_WEAK */
#endif
} actionnode;

typedef actionnode promptnode;

typedef struct marknode {
    struct marknode *next;
    char *pattern;
    int attrcode;
} marknode;

typedef struct keynode {
    struct keynode *next;	/* next key definition in list */
    char *name;			/* key name */
    char *sequence;		/* escape sequence sent by terminal */
    int seqlen;			/* GH: length of esc seq to allow \0 in seq */
    void (*funct)();		/* function called when key pressed */
    char *call_data;		/* data passed to function */
} keynode;

typedef struct delaynode {
    struct delaynode *next;
    char *name;
    char *command;
    vtime when;                 /* structure containing time when */
				/* command must be executed */
} delaynode;

typedef struct varnode {        /* for named variables */
    struct varnode *next;
    char *name;
    int index;
    long num;
    char *str;
} varnode;

typedef struct {		/* GH: for keeping track of color status */
    char *fg, *bg;
    int uline, bold, standout, inv, blink;
} colorstatus;

/* Variable struct definitions */
typedef struct {                       /* holds value of local parameters */
  long num[NUMPARAM];
  char str[NUMPARAM][PARAMLEN];
} params;

typedef struct {                       /* vector element of pointers */
  long *num;                           /* to ALL variables */
  char *str;
} vars;

typedef struct {
  params *p[MAX_STACK];
  int curr;
} param_stack;

#define LM_NOECHO 1	/* no local echo */
#define LM_CHAR 2	/* char-by-char mode (no line editing) */

/* editing session control */
typedef struct editsess {
  int key;		/* session identifier (-1 if cancelled) */
  char builtin;         /* 1 if viewing built-in text */
  int pid;		/* pid of child */
  char *descr;  	/* short description of what we are editing */
  char *file;		/* name of temporary file */
  time_t ctime; 	/* time when temp file was created (upper bound) */
  long oldsize; 	/* original file size */
  struct editsess *next;
} editsess;

/* shared vars from powwow.c */
extern char identified;
extern int internal;
extern char history_done;
extern int linemode;		/* line mode with flags as above (LM_*) */
extern char hostname[], portnumber[];
extern char deffile[], helpfile[];

extern int cols, lines;		/* terminal window size */
extern int line0, col0;		/* origin of input line */
extern char promptstr[];	/* prompt string */
extern int promptlen;           /* length of prompt string */
extern char surely_isprompt;    /* 1 if #prompt set #isprompt */
extern char edbuf[];		/* input line buffer */
extern int edlen;		/* characters in edbuf */
extern int pos;			/* cursor position in edbuf */
extern char edattrbeg[], edattrend[];

extern int need_prompt_redraw;	/* GH: set to true when window size changes */
extern int window_size_changed; /* GH: set to true to send new win size */

extern long received, sent;

extern clock_t start_clock, cpu_clock;

extern aliasnode *aliases;
extern actionnode *actions;
extern promptnode *prompts;
extern marknode *markers;
extern int a_nice;
extern keynode *keydefs;
extern delaynode *delays;
extern delaynode *dead_delays;
extern varnode *named_vars[2];
extern int num_named_vars[2];
extern int max_named_vars;
extern vars *var;
extern vtime now, start_time, ref_time;

extern char initstr[];
extern FILE *capfile, *recordfile;
extern char prefixstr[];
extern char inserted_next[];
extern char verbatim;

extern char opt_exit;
extern char opt_history;
extern char opt_words;
extern char opt_compact;
extern char opt_debug;
extern char opt_wrap;
extern char echo_ext;
extern char echo_int;
extern char echo_key;
extern char opt_speedwalk;
extern char opt_autoprint;

extern char *hist[];
extern int curline;
extern int pickline;

extern void (*last_edit_cmd)();

extern char *delimeters[];
extern char *delim_mode_names[];
extern int word_delim_mode;

extern char action_chars[];

/* shared functions from powwow.c */

void set_deffile();
int read_settings();
int save_settings();
void signal_init();
void exitpowwow();
void execute();
void syserr();
void errmsg();
char *skipspace();
void process_remote_input();
void unescape();
void put_marks();
void wrap_print();
void smart_print();
void suspend_powwow();
void send_to_main_host();
void setfile();
void strdel();
void strins();
void showprompt();
void parse_user_input();
char *parse_instruction();
void to_input_line();
void cancel_edit();
void set_param();
void restore_param();
void subst_param();
void printver();
void flashparen();
void put_in_history();
char *split_first_word();
int printstrlen();
int parse_walk();
char *str_dup();
void put_word();
void set_custom_delimeters();
void (*sig_permanent())();
int is_permanent_variable();

#endif /* _POWWOW_H_ */
